@@ -3,7 +3,8 @@
 #include "timers.h"
 #include "device.h"
 #include "peer.h"
-#include "packets.h"
+#include "queueing.h"
+#include "socket.h"
 
 /*
  * Timer for retransmitting the handshake if we don't hear back after `REKEY_TIMEOUT + jitter` ms
@@ -28,24 +29,26 @@ static void expired_retransmit_handshake(unsigned long ptr)
 {
 	peer_get_from_ptr(ptr);
 	if (peer->timer_handshake_attempts > MAX_TIMER_HANDSHAKES) {
-		pr_debug("%s: Handshake for peer %Lu (%pISpfsc) did not complete after %d attempts, giving up\n", netdev_pub(peer->device)->name, peer->internal_id, &peer->endpoint.addr, MAX_TIMER_HANDSHAKES + 2);
+		pr_debug("%s: Handshake for peer %Lu (%pISpfsc) did not complete after %d attempts, giving up\n", peer->device->dev->name, peer->internal_id, &peer->endpoint.addr, MAX_TIMER_HANDSHAKES + 2);
 
-		del_timer(&peer->timer_send_keepalive);
-		/* We remove all existing packets and don't try again,
+		if (likely(peer->timers_enabled))
+			del_timer(&peer->timer_send_keepalive);
+		/* We drop all packets without a keypair and don't try again,
 		 * if we try unsuccessfully for too long to make a handshake. */
-		skb_queue_purge(&peer->tx_packet_queue);
+		skb_queue_purge(&peer->staged_packet_queue);
+
 		/* We set a timer for destroying any residue that might be left
 		 * of a partial exchange. */
-		if (likely(peer->timers_enabled))
-			mod_timer(&peer->timer_kill_ephemerals, jiffies + (REJECT_AFTER_TIME * 3));
+		if (likely(peer->timers_enabled) && !timer_pending(&peer->timer_zero_key_material))
+			mod_timer(&peer->timer_zero_key_material, jiffies + (REJECT_AFTER_TIME * 3));
 	} else {
 		++peer->timer_handshake_attempts;
-		pr_debug("%s: Handshake for peer %Lu (%pISpfsc) did not complete after %d seconds, retrying (try %d)\n", netdev_pub(peer->device)->name, peer->internal_id, &peer->endpoint.addr, REKEY_TIMEOUT / HZ, peer->timer_handshake_attempts + 1);
+		pr_debug("%s: Handshake for peer %Lu (%pISpfsc) did not complete after %d seconds, retrying (try %d)\n", peer->device->dev->name, peer->internal_id, &peer->endpoint.addr, REKEY_TIMEOUT / HZ, peer->timer_handshake_attempts + 1);
 
 		/* We clear the endpoint address src address, in case this is the cause of trouble. */
 		socket_clear_peer_endpoint_src(peer);
 
-		packet_queue_handshake_initiation(peer, true);
+		packet_send_queued_handshake_initiation(peer, true);
 	}
 	peer_put(peer);
 }
@@ -56,7 +59,7 @@ static void expired_send_keepalive(unsigned long ptr)
 	packet_send_keepalive(peer);
 	if (peer->timer_need_another_keepalive) {
 		peer->timer_need_another_keepalive = false;
-		if (peer->timers_enabled)
+		if (likely(peer->timers_enabled))
 			mod_timer(&peer->timer_send_keepalive, jiffies + KEEPALIVE_TIMEOUT);
 	}
 	peer_put(peer);
@@ -65,23 +68,23 @@ static void expired_send_keepalive(unsigned long ptr)
 static void expired_new_handshake(unsigned long ptr)
 {
 	peer_get_from_ptr(ptr);
-	pr_debug("%s: Retrying handshake with peer %Lu (%pISpfsc) because we stopped hearing back after %d seconds\n", netdev_pub(peer->device)->name, peer->internal_id, &peer->endpoint.addr, (KEEPALIVE_TIMEOUT + REKEY_TIMEOUT) / HZ);
+	pr_debug("%s: Retrying handshake with peer %Lu (%pISpfsc) because we stopped hearing back after %d seconds\n", peer->device->dev->name, peer->internal_id, &peer->endpoint.addr, (KEEPALIVE_TIMEOUT + REKEY_TIMEOUT) / HZ);
 	/* We clear the endpoint address src address, in case this is the cause of trouble. */
 	socket_clear_peer_endpoint_src(peer);
-	packet_queue_handshake_initiation(peer, false);
+	packet_send_queued_handshake_initiation(peer, false);
 	peer_put(peer);
 }
 
-static void expired_kill_ephemerals(unsigned long ptr)
+static void expired_zero_key_material(unsigned long ptr)
 {
 	peer_get_from_ptr(ptr);
-	if (!queue_work(peer->device->peer_wq, &peer->clear_peer_work)) /* Takes our reference. */
+	if (!queue_work(peer->device->handshake_send_wq, &peer->clear_peer_work)) /* Takes our reference. */
 		peer_put(peer); /* If the work was already on the queue, we want to drop the extra reference */
 }
-static void queued_expired_kill_ephemerals(struct work_struct *work)
+static void queued_expired_zero_key_material(struct work_struct *work)
 {
 	struct wireguard_peer *peer = container_of(work, struct wireguard_peer, clear_peer_work);
-	pr_debug("%s: Zeroing out all keys for peer %Lu (%pISpfsc), since we haven't received a new one in %d seconds\n", netdev_pub(peer->device)->name, peer->internal_id, &peer->endpoint.addr, (REJECT_AFTER_TIME * 3) / HZ);
+	pr_debug("%s: Zeroing out all keys for peer %Lu (%pISpfsc), since we haven't received a new one in %d seconds\n", peer->device->dev->name, peer->internal_id, &peer->endpoint.addr, (REJECT_AFTER_TIME * 3) / HZ);
 	noise_handshake_clear(&peer->handshake);
 	noise_keypairs_clear(&peer->keypairs);
 	peer_put(peer);
@@ -90,8 +93,11 @@ static void queued_expired_kill_ephemerals(struct work_struct *work)
 static void expired_send_persistent_keepalive(unsigned long ptr)
 {
 	peer_get_from_ptr(ptr);
-	if (likely(peer->persistent_keepalive_interval))
+	if (likely(peer->persistent_keepalive_interval)) {
+		if (likely(peer->timers_enabled))
+			del_timer(&peer->timer_send_keepalive);
 		packet_send_keepalive(peer);
+	}
 	peer_put(peer);
 }
 
@@ -136,14 +142,15 @@ void timers_handshake_complete(struct wireguard_peer *peer)
 	if (likely(peer->timers_enabled))
 		del_timer(&peer->timer_retransmit_handshake);
 	peer->timer_handshake_attempts = 0;
+	peer->sent_lastminute_handshake = false;
 	do_gettimeofday(&peer->walltime_last_handshake);
 }
 
 /* Should be called after an ephemeral key is created, which is before sending a handshake response or after receiving a handshake response. */
-void timers_ephemeral_key_created(struct wireguard_peer *peer)
+void timers_session_derived(struct wireguard_peer *peer)
 {
 	if (likely(peer->timers_enabled))
-		mod_timer(&peer->timer_kill_ephemerals, jiffies + (REJECT_AFTER_TIME * 3));
+		mod_timer(&peer->timer_zero_key_material, jiffies + (REJECT_AFTER_TIME * 3));
 }
 
 /* Should be called before a packet with authentication -- data, keepalive, either handshake -- is sent, or after one is received. */
@@ -159,9 +166,9 @@ void timers_init_peer(struct wireguard_peer *peer)
 	setup_timer(&peer->timer_retransmit_handshake, expired_retransmit_handshake, (unsigned long)peer);
 	setup_timer(&peer->timer_send_keepalive, expired_send_keepalive, (unsigned long)peer);
 	setup_timer(&peer->timer_new_handshake, expired_new_handshake, (unsigned long)peer);
-	setup_timer(&peer->timer_kill_ephemerals, expired_kill_ephemerals, (unsigned long)peer);
+	setup_timer(&peer->timer_zero_key_material, expired_zero_key_material, (unsigned long)peer);
 	setup_timer(&peer->timer_persistent_keepalive, expired_send_persistent_keepalive, (unsigned long)peer);
-	INIT_WORK(&peer->clear_peer_work, queued_expired_kill_ephemerals);
+	INIT_WORK(&peer->clear_peer_work, queued_expired_zero_key_material);
 }
 
 void timers_uninit_peer(struct wireguard_peer *peer)
@@ -173,7 +180,7 @@ void timers_uninit_peer(struct wireguard_peer *peer)
 	del_timer_sync(&peer->timer_retransmit_handshake);
 	del_timer_sync(&peer->timer_send_keepalive);
 	del_timer_sync(&peer->timer_new_handshake);
-	del_timer_sync(&peer->timer_kill_ephemerals);
+	del_timer_sync(&peer->timer_zero_key_material);
 	del_timer_sync(&peer->timer_persistent_keepalive);
 	flush_work(&peer->clear_peer_work);
 }
